﻿#include <queue>
#include <mutex>
#include <condition_variable>

class blocking_bounded_queue {
public:
    blocking_bounded_queue(std::size_t size = 10): size_limit{ size }, closing{ false } {}

    bool add(int e) {
        std::unique_lock<std::mutex> lg(mutex);
        add_variable.wait(lg, [&]() -> bool { return queue.size() < size_limit || closing; });
        if (closing) {
            return false;
        }
        queue.push(e);
        take_variable.notify_one();
        return true;
    }

    bool take(int& out) {
        std::unique_lock<std::mutex> lg(mutex);
        take_variable.wait(lg, [&]() -> bool { return queue.size() > 0 || closing; });
        if (queue.size() == 0) {
            return false;
        }
        out = queue.front(); queue.pop();
        add_variable.notify_one();
        return true;
    }

    void close() {
        std::unique_lock<std::mutex> lg(mutex);
        closing = true;
        add_variable.notify_all();
        take_variable.notify_all();
    }

    ~blocking_bounded_queue() {
        close();
    }

private:
    std::queue<int> queue;
    std::size_t size_limit;
    std::mutex mutex;
    std::condition_variable take_variable, add_variable;
    bool closing;
};


#include <iostream>

int main() {
    blocking_bounded_queue bbq;
    auto prod = [&]() {
        for (int i = 0; i < 1'000; ++i) {
            bbq.add(i);
        }
        bbq.close();
    };

    auto cons = [&]() {
        int out;
        while (bbq.take(out)) {
            std::cout << out << '\n';
        }
    };

    std::thread t1(prod);
    std::thread t2(cons);

    t1.join();
    t2.join();

}
